local band = bit.band
local tinsert = table.insert
local tremove = table.remove

local Addon = CreateFrame("Frame", "BossSwingTimer")

local AddonPath = "Interface\\AddOns\\BossSwingTimer\\"

local function MergeTables(dst, src)
  for k, v in pairs (src) do
    if type (v) ~= "table" then
      if dst[k] == nil then
        dst[k] = v
      end
    else
      if type(dst[k]) ~= "table" then
        dst[k] = {}
      end
      MergeTables(dst[k], v)
    end
  end
end
local DefaultDB = {
  point = {
    x = 0,
    y = 0,
    p = "CENTER",
    r = "CENTER",
  },
  width = 300,
  height = 32,
  length = 2.5,
  speeds = {},
  show = true,
}

Addon.swings = {}

local unitList = {{id = "target", color = {r = 1, g = 0, b = 0}}, {id = "focus", color = {r = 1, g = 1, b = 0}}}

local function npcid(guid)
  return tonumber("0x" .. guid:sub(7, 10))
end
local function isnpc(guid)
  local t = band(tonumber("0x" .. guid:sub(5, 5)), 7)
  return t == 3 or t == 5
end

function Addon:CreateUI()
  self.bar = CreateFrame("Frame", "BossSwingTimerBar", UIParent)
  self.bar:SetMovable(true)
  self.bar:SetResizable(true)
  self.bar.lag = self.bar:CreateTexture(nil, "BACKGROUND")
  self.bar.lag:SetPoint("TOPLEFT", self.bar, "TOPLEFT", 0, 0)
  self.bar.lag:SetPoint("BOTTOMLEFT", self.bar, "BOTTOMLEFT", 0, 0)
  self.bar.lag:SetWidth(1)
  self.bar.lag:SetTexture(AddonPath .. "glaze.tga")
  self.bar.lag:SetVertexColor(0.6, 0.0, 0.0)
  self.bar.lag:SetAlpha(0.8)
  self.bar.background = self.bar:CreateTexture(nil, "BACKGROUND")
  self.bar.background:SetPoint("TOPLEFT", self.bar.lag, "TOPRIGHT", 0, 0)
  self.bar.background:SetPoint("BOTTOMRIGHT", self.bar, "BOTTOMRIGHT", 0, 0)
  self.bar.background:SetTexture(AddonPath .. "glaze.tga")
  self.bar.background:SetVertexColor(0.4, 0.4, 0.4)
  self.bar.background:SetAlpha(0.8)
  self.bar:EnableMouse(true)
  self.bar:SetScript("OnMouseDown", function() self.bar:StartMoving() end)
  self.bar:SetScript("OnMouseUp", function()
    self.bar:StopMovingOrSizing()
    self.db.point.p, _, self.db.point.r, self.db.point.x, self.db.point.y = self.bar:GetPoint()
  end)
  self.bar:SetScript("OnUpdate", function() self:OnUpdate() end)

  self.bar.grip = CreateFrame("Frame", nil, self.bar)
  self.bar.grip.tex = self.bar.grip:CreateTexture(nil, "OVERLAY")
  self.bar.grip:SetSize(10, 10)
  self.bar.grip:SetPoint("BOTTOMRIGHT", self.bar, "BOTTOMRIGHT", 0, 0)
  self.bar.grip:EnableMouse(true)
  self.bar.grip:SetScript("OnMouseDown", function() self.bar:StartSizing("BOTTOMRIGHT") end)
  self.bar.grip:SetScript("OnMouseUp", function()
    self.bar:StopMovingOrSizing()
    self.db.width = self.bar:GetWidth()
    self.db.height = self.bar:GetHeight()
  end)
  self.bar.grip.tex:SetAllPoints()
  self.bar.grip.tex:SetAlpha(0)
  self.bar.grip:Show()

  self.bar:SetSize(self.db.width, self.db.height)
  self.bar:SetPoint(self.db.point.p, UIParent, self.db.point.r, self.db.point.x, self.db.point.y)
  if self.db.show then
    self.bar:Show()
  end
end

function Addon:OnSwing(time, guid, damage)
  for _, u in ipairs(unitList) do
    local uid = UnitGUID(u.id)
    if uid and isnpc(uid) then
      local speed = UnitAttackSpeed(u.id)
      if speed then
        self.db.speeds[npcid(uid)] = {value = speed, api = true}
      end
    end
  end

  local id = npcid(guid)
  self.swings[guid] = self.swings[guid] or {}
  local prev = self.swings[guid].time
  local speed = nil
  if prev and (time - prev) < 5 then
    speed = time - prev
  end
  if self.db.speeds[id] and self.db.speeds[id].api then
    speed = self.db.speeds[id].value
  elseif speed then
    self.db.speeds[id] = {value = speed}
  end
  if speed then
    self.swings[guid].time = time + speed
    if damage then
      self.swings[guid].damage = damage
    end
  end
end

-- gui --

Addon.texpool = {}
function Addon:CreateTick()
  if #self.texpool > 0 then
    return tremove(self.texpool)
  end
  local result = self.bar:CreateTexture(nil, "ARTWORK")
  result:SetTexture(AddonPath .. "tick.tga")
  result:SetTexCoord(0.40625, 0.5625, 0, 1)
  result:SetPoint("TOP", self.bar, "TOP", 0, 0)
  result:SetPoint("BOTTOM", self.bar, "BOTTOM", 0, 0)
  result:SetWidth(5)
  return result
end
function Addon:RecycleTick(tex)
  tinsert(self.texpool, tex)
  tex:Hide()
end

function Addon:OnUpdate()
  local special = {}
  for _, u in ipairs(unitList) do
    local uid = UnitGUID(u.id)
    if uid then
      special[uid] = u.color
    end
  end

  local _, _, _, lag = GetNetStats()
  lag = lag / 1000
  if lag > self.db.length then
    lag = self.db.length
  end
  self.bar.lag:SetWidth(lag / self.db.length * self.db.width)

  local maxDamage = 1
  local time = GetTime()
  for k, v in pairs(self.swings) do
    if not v.time or v.time < time then
      if v.tick then
        self:RecycleTick(v.tick)
        v.tick = nil
      end
    elseif v.damage and v.damage > maxDamage then
      maxDamage = v.damage
    end
  end
  for k, v in pairs(self.swings) do
    if v.tick and v.time < time + self.db.length then
      if not v.tick then
        v.tick = self:CreateTick()
      end
      local alpha = 1
      local color = {r = 1, g = 1, b = 1}
      if special[k] then
        color = special[k]
      else
        if v.damage and v.damage < maxDamage then
          alpha = v.damage / maxDamage
        end
      end
      v.tick:SetVertexColor(color.r, color.g, color.b, alpha)
      v.tick:SetPoint("LEFT", self.bar, "LEFT", (v.time - time) / self.db.length * self.db.width - 2, 0)
      v.tick:Show()
    end 
  end
end

---------

Addon:SetScript("OnEvent", function(self, event, ...)
  if self[event] then
    self[event](self, ...)
  end
end)
Addon:RegisterEvent("ADDON_LOADED")

function Addon:ADDON_LOADED(addon)
  if addon == "BossSwingTimer" then
    BossSwingTimerDB = BossSwingTimerDB or {}
    self.db = BossSwingTimerDB
    MergeTables(self.db, DefaultDB)

    self:CreateUI()

    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

    SlashCmdList["BossSwingTimer"] = function (cmd) Addon:OnCommand (cmd) end
    SLASH_BossSwingTimer1 = "/bst"
  end
end

function Addon:COMBAT_LOG_EVENT_UNFILTERED(...)
  local time = GetTime()
  local event = select(2, ...)
  local guid = select(4, ...)
  if select(8, ...) == UnitGUID("player") and isnpc(guid) and (event == "SWING_DAMAGE" or event == "SWING_MISSED") then
    self:OnSwing(time, guid, event == "SWING_DAMAGE" and select(15, ...))
  end
end

function Addon:OnCommand(cmd)
  self.db.show = not self.db.show
  if self.db.show then
    self.bar:Show()
  else
    self.bar:Hide()
  end
end
